@@ -15,12 +15,22 @@
 #include "poa-private.h"
 #include "orbit-poa.h"
 
-#define ORBIT_SERVANT_SET_CLASSINFO(servant,ci) {                      \
-  ((PortableServer_ServantBase *)(servant))->vepv[0]->_private = (ci); \
-}
+#ifdef DEBUG_LOCKS
+#  define LOCK_DEBUG(a) g_printerr("%p: %6s file %s: line %d (%s)\n", \
+				   poa, a, __FILE__, __LINE__, __PRETTY_FUNCTION__);
+#else
+#  define LOCK_DEBUG(a)
+#endif
 
-#define POA_LOCK(poa)   LINK_MUTEX_LOCK(poa->base.lock)
-#define POA_UNLOCK(poa) LINK_MUTEX_UNLOCK(poa->base.lock)
+#define POA_LOCK(poa)   G_STMT_START { \
+				LOCK_DEBUG("lock") \
+				LINK_MUTEX_LOCK(poa->base.lock); \
+			} G_STMT_END
+
+#define POA_UNLOCK(poa) G_STMT_START { \
+				LINK_MUTEX_UNLOCK(poa->base.lock); \
+				LOCK_DEBUG("unlock") \
+			} G_STMT_END
 
 static GMutex     *ORBit_class_assignment_lock    = NULL;
 static GHashTable *ORBit_class_assignments        = NULL;
@@ -63,11 +73,93 @@ static void               ORBit_POAObject_handle_request    (ORBit_POAObject
 							     GIOPRecvBuffer    *recv_buffer,
 							     CORBA_Environment *ev);
 
-static void               ORBit_POA_deactivate_object       (PortableServer_POA  poa,
-							     ORBit_POAObject     pobj,
-							     CORBA_boolean       do_etherealize,
-							     CORBA_boolean       is_cleanup);
+static void               ORBit_POA_deactivate_object_T   (PortableServer_POA  poa,
+							   ORBit_POAObject     pobj,
+							   CORBA_boolean       do_etherealize,
+							   CORBA_boolean       is_cleanup);
      
+
+static CORBA_Object
+ORBit_POA_obj_to_ref (PortableServer_POA  poa,
+                      ORBit_POAObject     pobj,
+                      const CORBA_char   *intf,
+                      CORBA_Environment  *ev);
+
+static void
+ORBit_POA_invocation_stack_push (PortableServer_POA  poa,
+                                 ORBit_POAObject     pobj)
+{
+ 	LINK_MUTEX_LOCK (poa->orb->lock);
+	poa->orb->current_invocations =
+		g_slist_prepend (poa->orb->current_invocations, pobj);
+	LINK_MUTEX_UNLOCK (poa->orb->lock);
+}
+                                                                                
+static void
+ORBit_POA_invocation_stack_pop (PortableServer_POA  poa,
+                                ORBit_POAObject pobj)
+{
+	/* FIXME, we know it is the top element to be removed */ 
+	LINK_MUTEX_LOCK (poa->orb->lock);
+	poa->orb->current_invocations =
+		g_slist_remove (poa->orb->current_invocations, pobj);
+	LINK_MUTEX_UNLOCK (poa->orb->lock);
+}
+
+static  PortableServer_ObjectId*
+ORBit_POA_invocation_stack_lookup_objid (PortableServer_POA          poa,
+                                         PortableServer_ServantBase *servant)
+{
+	PortableServer_ObjectId *objid = NULL;
+	GSList *l;
+
+	LINK_MUTEX_LOCK (poa->orb->lock);
+	for (l = poa->orb->current_invocations; l; l = l->next) {
+		ORBit_POAObject pobj = l->data;
+		
+		if (pobj->servant == servant)
+			objid = ORBit_sequence_CORBA_octet_dup (pobj->object_id);
+	}
+	LINK_MUTEX_UNLOCK (poa->orb->lock);
+
+	return objid;
+}
+
+static  CORBA_Object
+ORBit_POA_invocation_stack_lookup_objref (PortableServer_POA           poa,
+                                          PortableServer_ServantBase *servant)
+{
+	CORBA_Object  result = CORBA_OBJECT_NIL; 
+	GSList *l;
+
+	LINK_MUTEX_LOCK (poa->orb->lock);
+	for (l = poa->orb->current_invocations; l; l = l->next) {
+		ORBit_POAObject pobj = l->data;
+		
+		if (pobj->servant == servant)
+			result = ORBit_POA_obj_to_ref (poa, pobj, NULL, NULL);
+	}
+	LINK_MUTEX_UNLOCK (poa->orb->lock);
+
+	return result;
+}
+
+/* FIXME, I would prefer if argument would be of type PortableServer_POA */ 
+static  ORBit_POAObject
+ORBit_POA_invocation_stack_peek (CORBA_ORB        orb)
+{
+	ORBit_POAObject pobj = CORBA_OBJECT_NIL;
+
+	LINK_MUTEX_LOCK (orb->lock);
+	if (orb->current_invocations == NULL)
+		pobj = CORBA_OBJECT_NIL;
+	else
+		pobj = (ORBit_POAObject) orb->current_invocations->data;
+	LINK_MUTEX_UNLOCK (orb->lock);
+	
+	return pobj;
+}
+
 /* PortableServer_Current interface */
 static void
 ORBit_POACurrent_free_fn (ORBit_RootObject obj_in)
@@ -107,16 +199,12 @@ ORBit_POACurrent_get_object (PortableServer_Current  obj,
 	ORBit_POAObject pobj = CORBA_OBJECT_NIL;
 
 	g_assert (obj && obj->parent.interface->type == ORBIT_ROT_POACURRENT);
-
-	LINK_MUTEX_LOCK (obj->orb->lock);
-	if (obj->orb->current_invocations == NULL)
+	
+	if ( ! (pobj = ORBit_POA_invocation_stack_peek (obj->orb)))
 		CORBA_exception_set_system
 			(ev, ex_PortableServer_Current_NoContext,
 			 CORBA_COMPLETED_NO);
-	else
-		pobj = (ORBit_POAObject) obj->orb->current_invocations->data;
-	LINK_MUTEX_UNLOCK (obj->orb->lock);
-
+	
 	return pobj;
 }
 
@@ -154,6 +242,8 @@ ORBit_classinfo_register (PortableServer_ClassInfo *ci)
 			     (gpointer) ci->class_name, ci);
 }
 
+#define VEPV_CACHE_SIZE(c) (c)[0]
+
 void
 ORBit_skel_class_register (PortableServer_ClassInfo   *ci,
 			   PortableServer_ServantBase *servant,
@@ -177,10 +267,11 @@ ORBit_skel_class_register (PortableServer_ClassInfo   *ci,
 	ORBit_classinfo_register (ci);
 	if (!ci->vepvmap) {
 		CORBA_unsigned_long offset;
+		CORBA_unsigned_long epvmap_size;
 
-		/* FIXME: we could propagate the size nicely here */
-		ci->vepvmap = g_new0 (ORBit_VepvIdx, 
-				      *(ci->class_id) + 1);
+		epvmap_size = *(ci->class_id) + 1;
+		ci->vepvmap = g_new0 (ORBit_VepvIdx, epvmap_size);
+		VEPV_CACHE_SIZE(ci->vepvmap) = epvmap_size;
 
 		ci->vepvmap[*(ci->class_id)] = class_offset;
 		
@@ -357,9 +448,9 @@ ORBit_POA_remove_child (PortableServer_POA poa,
 }
 
 static gboolean
-ORBit_POA_destroy (PortableServer_POA  poa,
-		   CORBA_boolean       etherealize_objects,
-		   CORBA_Environment  *ev)
+ORBit_POA_destroy_T_R (PortableServer_POA  poa,
+		       CORBA_boolean       etherealize_objects,
+		       CORBA_Environment  *ev)
 {
 	GPtrArray *adaptors;
 	int        numobjs;
@@ -373,12 +464,13 @@ ORBit_POA_destroy (PortableServer_POA  poa,
 	if (poa->life_flags & (ORBit_LifeF_Deactivating|ORBit_LifeF_Destroying))
 		return FALSE;	/* recursion */
 
-	/* FIXME: this needs some hard-core thread safety work ... */
-
 	poa->life_flags |= ORBit_LifeF_Destroying;
 
 	adaptors = poa->orb->adaptors;
 
+	LINK_MUTEX_LOCK (ORBit_RootObject_lifecycle_lock);
+	POA_UNLOCK (poa);
+
 	/* Destroying the children is tricky, b/c they may die
 	 * while we are traversing. We traverse over the
 	 * ORB's global list (rather than poa->child_poas) 
@@ -386,10 +478,23 @@ ORBit_POA_destroy (PortableServer_POA  poa,
 	for (i = 0; i < adaptors->len; i++) {
 		PortableServer_POA cpoa = g_ptr_array_index (adaptors, i);
 
-		if (cpoa && cpoa->parent_poa == poa)
-			ORBit_POA_destroy (cpoa, etherealize_objects, ev);
+		if (cpoa && cpoa != poa) {
+			ORBit_RootObject_duplicate_T (cpoa);
+			LINK_MUTEX_UNLOCK (ORBit_RootObject_lifecycle_lock);
+			POA_LOCK (cpoa);
+
+			if (cpoa->parent_poa == poa) 
+				ORBit_POA_destroy_T_R (cpoa, etherealize_objects, ev);
+
+			POA_UNLOCK (cpoa);
+			LINK_MUTEX_LOCK (ORBit_RootObject_lifecycle_lock);
+			ORBit_RootObject_release_T (cpoa);
+		}
 	}
 
+	POA_LOCK (poa);
+	LINK_MUTEX_UNLOCK (ORBit_RootObject_lifecycle_lock);
+
 	poa->default_servant = NULL;
 
 	if (g_hash_table_size (poa->child_poas) > 0 || poa->use_cnt ||
@@ -433,7 +538,7 @@ traverse_cb (PortableServer_ObjectId *oid,
 	if (pobj->use_cnt > 0)
 		info->in_use = TRUE;
 
-	ORBit_POA_deactivate_object (info->poa, pobj, info->do_etherealize, TRUE);
+	ORBit_POA_deactivate_object_T (info->poa, pobj, info->do_etherealize, TRUE);
 }
 
 static gboolean
@@ -779,7 +884,9 @@ ORBit_POA_setup_root (CORBA_ORB orb, CORBA_Environment *ev)
 {
 	PortableServer_POA poa;
 	CORBA_Policy       policybuf[1];
-	CORBA_PolicyList   policies = {1, 1, (CORBA_Object *) policybuf, CORBA_FALSE};
+	CORBA_PolicyList   policies = {1, 1, NULL, CORBA_FALSE};
+
+	policies._buffer = (CORBA_Object *) policybuf;
 
 	policies._buffer [0] = (CORBA_Policy)
 		PortableServer_POA_create_implicit_activation_policy (
@@ -937,10 +1044,10 @@ ORBit_POA_activate_object_T (PortableServer_POA          poa,
  * etherialization and memory release will occur later.
  */
 static void
-ORBit_POA_deactivate_object (PortableServer_POA poa,
-			     ORBit_POAObject    pobj,
-			     CORBA_boolean      do_etherealize,
-			     CORBA_boolean      is_cleanup)
+ORBit_POA_deactivate_object_T (PortableServer_POA poa,
+			       ORBit_POAObject    pobj,
+			       CORBA_boolean      do_etherealize,
+			       CORBA_boolean      is_cleanup)
 {
 	PortableServer_ServantBase *servant = pobj->servant;
 
@@ -997,11 +1104,18 @@ ORBit_POA_deactivate_object (PortableServer_POA poa,
 			/* In theory, the finalize fnc should always be non-NULL;
 			 * however, for backward compat. and general extended
 			 * applications we dont insist on it. */
-			if (epv && epv->finalize)
+			if (epv && epv->finalize) {
+				POA_UNLOCK (poa);
 				epv->finalize (servant, ev);
+				POA_LOCK (poa);
+			}
 		}
 		pobj->use_cnt--; /* allow re-activation */
-		g_assert (ev->_major == 0);
+
+		if (ev->_major != 0) {
+		  g_error ("finalize function for object %p threw an exception (%s). This is not allowed.",
+			   pobj, CORBA_exception_id (ev));
+		}
 
 		CORBA_exception_free (ev);
 	}
@@ -1107,8 +1221,10 @@ ORBit_POAObject_handle_request (ORBit_POAObject    pobj,
 	ORBitSmallSkeleton                   small_skel = NULL;
 	gpointer                             imp = NULL;
 
-	if (poa)
+	if (poa) {
+		ORBit_RootObject_duplicate (poa);
 		POA_LOCK (poa);
+	}
 
 	if (!poa || !poa->poa_manager)
 		CORBA_exception_set_system (
@@ -1124,9 +1240,7 @@ ORBit_POAObject_handle_request (ORBit_POAObject    pobj,
 						
 				poa->held_requests = g_slist_prepend (
 					poa->held_requests, recv_buffer);
-				
-				POA_UNLOCK (poa);
-				return;
+				goto clean_out;
 			} else
 				CORBA_exception_set_system (
 					ev, ex_CORBA_TRANSIENT,
@@ -1184,12 +1298,6 @@ ORBit_POAObject_handle_request (ORBit_POAObject    pobj,
 		}
 	}
 
-	pobj->use_cnt++;
-	LINK_MUTEX_LOCK (poa->orb->lock);
-	poa->orb->current_invocations =
-		g_slist_prepend (poa->orb->current_invocations, pobj);
-	LINK_MUTEX_UNLOCK (poa->orb->lock);
-	
 	if (ev->_major == CORBA_NO_EXCEPTION && !pobj->servant)
 		CORBA_exception_set_system (
 			ev, ex_CORBA_OBJECT_NOT_EXIST, 
@@ -1200,6 +1308,9 @@ ORBit_POAObject_handle_request (ORBit_POAObject    pobj,
 		goto clean_out;
 	}
 
+	pobj->use_cnt++;
+	ORBit_POA_invocation_stack_push (poa, pobj);
+	
 	klass = ORBIT_SERVANT_TO_CLASSINFO (pobj->servant);
 
 	if (klass->impl_finder)
@@ -1228,30 +1339,26 @@ ORBit_POAObject_handle_request (ORBit_POAObject    pobj,
 		}
 	}
 
-	if (ev->_major != CORBA_NO_EXCEPTION) {
+	if (ev->_major != CORBA_NO_EXCEPTION)
 		return_exception (recv_buffer, m_data, ev);
-		goto clean_out;
+
+	else {
+		POA_UNLOCK (poa);
+
+		if (recv_buffer) {
+			struct ORBit_POA_invoke_data invoke_data;
+			
+			invoke_data.small_skel = small_skel;
+			invoke_data.imp        = imp;
+			
+			ORBit_small_invoke_adaptor ((ORBit_OAObject) pobj, recv_buffer,
+						    m_data, &invoke_data, ev);
+		} else
+			small_skel (pobj->servant, ret, args, ctx, ev, imp);
+		
+		POA_LOCK (poa);
 	}
 
-	ORBit_RootObject_duplicate (poa);
-	POA_UNLOCK (poa);
-
-	if (recv_buffer) {
-		struct ORBit_POA_invoke_data invoke_data;
-
-		invoke_data.small_skel = small_skel;
-		invoke_data.imp        = imp;
-
-		ORBit_small_invoke_adaptor (
-			(ORBit_OAObject) pobj, recv_buffer,
-			m_data, &invoke_data, ev);
-	} else
-		small_skel (pobj->servant, ret, args, ctx, ev, imp);
-
-	POA_LOCK (poa);
-	ORBit_RootObject_release (poa);
-
- clean_out:
 	if (recv_buffer)
 		CORBA_exception_free (ev);
 
@@ -1263,24 +1370,23 @@ ORBit_POAObject_handle_request (ORBit_POAObject    pobj,
 				oid, pobj->servant, ev);
 			break;
 		case PortableServer_USE_DEFAULT_SERVANT:
-			ORBit_POA_deactivate_object (poa, pobj, FALSE, FALSE);
+			ORBit_POA_deactivate_object_T (poa, pobj, FALSE, FALSE);
 			break;
 		default:
 			g_assert_not_reached ();
 			break;
 		}
 
-	LINK_MUTEX_LOCK (poa->orb->lock);
-/*	g_assert ((ORBit_POAObject)poa->orb->current_invocations->data == pobj); */
-	poa->orb->current_invocations =
-		g_slist_remove (poa->orb->current_invocations, pobj);
-	LINK_MUTEX_UNLOCK (poa->orb->lock);
+	ORBit_POA_invocation_stack_pop (poa, pobj);
+
 	pobj->use_cnt--;
 
 	if (pobj->life_flags & ORBit_LifeF_NeedPostInvoke)
 		ORBit_POAObject_post_invoke (pobj);             
 
+ clean_out:
 	POA_UNLOCK (poa);
+	ORBit_RootObject_release (poa);
 }
 
 static ORBit_POAObject
@@ -1364,23 +1470,41 @@ poa_invoke_at_idle (gpointer data)
 	ORBit_POAObject_invoke_incoming_request
 		(pcl->pobj, pcl->recv_buffer, NULL);
 
-	g_free (data);
+	pcl->pobj = NULL;
+	pcl->recv_buffer = NULL;
 
 	return FALSE;
 }
 
+static void
+poa_destroy_idle_closure (gpointer data)
+{
+	PoaIdleClosure *pcl = data;
+
+	if (pcl->pobj)
+		ORBit_RootObject_release (pcl->pobj);
+	if (pcl->recv_buffer)
+		giop_recv_buffer_unuse (pcl->recv_buffer);
+
+	g_free (pcl);
+}
+
 static void
 push_request_idle (ORBit_POAObject  *pobj,
-		   GIOPRecvBuffer  **recv_buffer)
+		   GIOPRecvBuffer  **recv_buffer,
+		   GMainContext     *on_context)
 {
 	PoaIdleClosure *pcl = g_new (PoaIdleClosure, 1);
+	GSource *source;
 
 	pcl->pobj = *pobj;
 	pcl->recv_buffer = *recv_buffer;
 
-	/* FIXME: track pending idle impls
-	 * for cleanup at shutdown ... */
-	g_idle_add (poa_invoke_at_idle, pcl);
+	source = g_idle_source_new ();
+	g_source_set_callback (source, poa_invoke_at_idle, pcl, poa_destroy_idle_closure);
+
+	g_source_attach (source, on_context);
+	g_source_unref (source);
 	
 	*pobj = NULL;
 	*recv_buffer = NULL;
@@ -1461,6 +1585,10 @@ ORBit_POA_handle_request (PortableServer_POA poa,
 				
 				switch (adaptor->thread_hint) {
 				case ORBIT_THREAD_HINT_PER_OBJECT:
+					/* Note: If the user bound this object to a specific
+					 * thread this thread will be called, otherwise a
+					 * new thread will be created for handling requests
+					 * to the object. */
 					pool_push_request_for_T (pobj, &pobj, &recv_buffer);
 					break;
 
@@ -1483,7 +1611,11 @@ ORBit_POA_handle_request (PortableServer_POA poa,
 								&pobj, &recv_buffer);
 					/* drop through */
 				case ORBIT_THREAD_HINT_ALL_AT_IDLE:
-					push_request_idle (&pobj, &recv_buffer);
+					push_request_idle (&pobj, &recv_buffer, NULL);
+					break;
+
+				case ORBIT_THREAD_HINT_ON_CONTEXT:
+					push_request_idle (&pobj, &recv_buffer, adaptor->context);
 					break;
 
 				case ORBIT_THREAD_HINT_NONE:
@@ -1620,8 +1752,7 @@ ORBit_POA_ServantManager_unuse_servant (PortableServer_POA                    po
 }
 
 /*
- * This function is invoked from the generated stub code, after
- * invoking the servant method.
+ * Was this ever exposed / can we axe it ?
  */
 void
 ORBit_POAObject_post_invoke (ORBit_POAObject pobj)
@@ -1634,7 +1765,7 @@ ORBit_POAObject_post_invoke (ORBit_POAObject pobj)
 		 * are stored in pobj->life_flags and they dont need
 		 * to be passed in again!
 		 */
-		ORBit_POA_deactivate_object (
+		ORBit_POA_deactivate_object_T (
 			pobj->poa, pobj, /*ether*/0, /*cleanup*/0);
 
 		/* WATCHOUT: pobj may not exist anymore! */
@@ -1817,18 +1948,23 @@ PortableServer_POA_destroy (PortableServer_POA   poa,
 
 	poa_sys_exception_if_fail (poa != NULL, ex_CORBA_INV_OBJREF);
 
-	if (poa->life_flags & ORBit_LifeF_Destroyed)
-		return;
+	ORBit_RootObject_duplicate (poa);
+	POA_LOCK (poa);
 
-	if (wait_for_completion && ORBit_POA_is_inuse (poa, CORBA_TRUE, ev)) {
+	if (poa->life_flags & ORBit_LifeF_Destroyed)
+		;
+
+	else if (wait_for_completion && ORBit_POA_is_inuse_T (poa, CORBA_TRUE, ev))
 		CORBA_exception_set_system (ev, ex_CORBA_BAD_INV_ORDER,
 					    CORBA_COMPLETED_NO);
-		return;
+	else {
+		done = ORBit_POA_destroy_T_R (poa, etherealize_objects, ev);
+
+		g_assert (done || !wait_for_completion);
 	}
 
-	done = ORBit_POA_destroy (poa, etherealize_objects, ev);
-
-	g_assert (done || !wait_for_completion);
+	POA_UNLOCK (poa);
+	ORBit_RootObject_release (poa);
 }
 
 CORBA_string
@@ -2047,7 +2183,7 @@ PortableServer_POA_deactivate_object (PortableServer_POA             poa,
 	pobj = ORBit_POA_object_id_lookup_T (poa, oid);
 
 	if (pobj && pobj->servant)
-		ORBit_POA_deactivate_object (poa, pobj, CORBA_TRUE, CORBA_FALSE);
+		ORBit_POA_deactivate_object_T (poa, pobj, CORBA_TRUE, CORBA_FALSE);
 
 	POA_UNLOCK (poa);
 
@@ -2139,7 +2275,6 @@ PortableServer_POA_servant_to_id (PortableServer_POA            poa,
 		objid = ORBit_sequence_CORBA_octet_dup (pobj->object_id);
 
 	} else {
-		GSList *l;
 		/*
 		 * FIXME:
 		 * This handles case 3 of the spec; but is broader:
@@ -2148,14 +2283,9 @@ PortableServer_POA_servant_to_id (PortableServer_POA            poa,
 		 * The stricter form could be implemented, 
 		 * but it would only add more code...
 		 */
-		LINK_MUTEX_LOCK (poa->orb->lock);
-		for (l = poa->orb->current_invocations; l; l = l->next) {
-			ORBit_POAObject pobj = l->data;
+		
+		objid = ORBit_POA_invocation_stack_lookup_objid (poa, servant);
 
-			if (pobj->servant == servant)
-				objid = ORBit_sequence_CORBA_octet_dup (pobj->object_id);
-		}
-		LINK_MUTEX_UNLOCK (poa->orb->lock);
 	}
 
 	if (!objid)
@@ -2203,7 +2333,6 @@ PortableServer_POA_servant_to_reference (PortableServer_POA            poa,
 
 		result = ORBit_POA_obj_to_ref (poa, pobj, NULL, ev);
 	} else {
-		GSList *l;
 		/*
 		 * FIXME:
 		 * This case deals with "invoked in the context of
@@ -2213,14 +2342,10 @@ PortableServer_POA_servant_to_reference (PortableServer_POA            poa,
 		 * go backward from servant to pobj to use_cnt, but we
 		 * dont do this since forward search is more general 
 		 */
-		LINK_MUTEX_LOCK (poa->orb->lock);
-		for (l = poa->orb->current_invocations; l; l = l->next) {
-			ORBit_POAObject pobj = l->data;
 
-			if (pobj->servant == servant)
-				result = ORBit_POA_obj_to_ref (poa, pobj, NULL, ev);
-		}
-		LINK_MUTEX_UNLOCK (poa->orb->lock);
+		
+		result = ORBit_POA_invocation_stack_lookup_objref (poa, servant);
+
 	}
 
 	if (result == CORBA_OBJECT_NIL)
@@ -2367,7 +2492,9 @@ ORBit_poa_init (void)
 }
 
 gboolean
-ORBit_poa_allow_cross_thread_call (ORBit_POAObject pobj)
+ORBit_poa_allow_cross_thread_call (ORBit_POAObject   pobj,
+				   ORBit_IMethodFlag method_flags)
+
 {
 	gpointer key = NULL;
 	GIOPThread *self;
@@ -2384,7 +2511,10 @@ ORBit_poa_allow_cross_thread_call (ORBit_POAObject pobj)
 
 	case PortableServer_ORB_CTRL_MODEL: {
 		ORBit_ObjectAdaptor adaptor = (ORBit_ObjectAdaptor) poa;
-				
+
+		if (method_flags & ORBit_I_METHOD_1_WAY)
+			return FALSE;
+
 		switch (adaptor->thread_hint) {
 		case ORBIT_THREAD_HINT_PER_OBJECT:
 			key = pobj;
@@ -2402,6 +2532,9 @@ ORBit_poa_allow_cross_thread_call (ORBit_POAObject pobj)
 			
 		case ORBIT_THREAD_HINT_ONEWAY_AT_IDLE:
 		case ORBIT_THREAD_HINT_ALL_AT_IDLE:
+		case ORBIT_THREAD_HINT_ON_CONTEXT:
+			/* FIXME: need GThread *g_main_context_get_owner()
+			   to do this right ... */
 		case ORBIT_THREAD_HINT_NONE:
 			break;
 		}
@@ -2421,20 +2554,22 @@ static gpointer
 get_c_method (CORBA_Object                 obj,
 	      glong                        class_id,
 	      PortableServer_ServantBase **servant,
-	      glong                        method_offset)
+	      glong                        method_offset,
+	      ORBit_IMethodFlag            method_flags)
 {
 	guchar *epv_start;
 	ORBit_POAObject pobj;
 
 	if (!obj ||
 	    !(pobj = (ORBit_POAObject) obj->adaptor_obj) ||
+	    !(pobj->base.interface->adaptor_type & ORBIT_ADAPTOR_POA) ||
 	    !(*servant = (PortableServer_ServantBase *) pobj->servant))
 		return NULL;
 
 	if (method_offset <= 0 || class_id <= 0)
 		return NULL;
 
-	if (!ORBit_poa_allow_cross_thread_call (pobj))
+	if (!ORBit_poa_allow_cross_thread_call (pobj, method_flags))
 		return NULL;
 
 	if (ORBit_small_flags & ORBIT_SMALL_FORCE_GENERIC_MARSHAL)
@@ -2444,8 +2579,8 @@ get_c_method (CORBA_Object                 obj,
 	 * FIXME: we could propagate the size of the vepvmap_cache
 	 * to here, and check it to avoid some really bogus stuff.
 	 */
-	if (!class_id || !(pobj->base.interface->adaptor_type & ORBIT_ADAPTOR_POA) ||
-	    !pobj->vepvmap_cache)
+	if (!class_id || !pobj->vepvmap_cache ||
+	    class_id >= VEPV_CACHE_SIZE (pobj->vepvmap_cache) )
 		return NULL;
 
 	epv_start = (guchar *) (*servant)->vepv [ pobj->vepvmap_cache [class_id] ];
@@ -2472,26 +2607,39 @@ ORBit_c_stub_invoke (CORBA_Object        obj,
 	gpointer method_impl;
 	PortableServer_ServantBase *servant;
 
+	if (method_index < 0 || method_index > methods->_length) {
+		CORBA_exception_set_system (ev, ex_CORBA_NO_IMPLEMENT,
+					    CORBA_COMPLETED_NO);
+		return;
+	}
+
 	if (skel_impl &&
 	    (method_impl = get_c_method (obj, class_id,
-					 &servant, method_offset))) {
+					 &servant, method_offset,
+					 methods->_buffer[method_index].flags))) {
 		
 		/* Unwound PreCall
-		   ++( ((ORBit_POAObject)(obj)->adaptor_obj)->use_cnt );	    \
-		   (obj)->orb->current_invocations =                           \
-		   g_slist_prepend ((obj)->orb->current_invocations,   \
-		   (obj)->adaptor_obj);               \
+		   POA_LOCK (((ORBit_POAObject)(obj)->adaptor_obj)->poa);
+		   ++( ((ORBit_POAObject)(obj)->adaptor_obj)->use_cnt );
+		   POA_UNLOCK (((ORBit_POAObject)(obj)->adaptor_obj)->poa);
+		   (obj)->orb->current_invocations =
+		   g_slist_prepend ((obj)->orb->current_invocations,
+		   (obj)->adaptor_obj);
 		*/
-
+		
+		CORBA_exception_init (ev);
 		skel_impl (servant, ret, args, ctx, ev, method_impl);
 
 		/* Unwound PostCall
-		   (obj)->orb->current_invocations =                                                      \
-		   g_slist_remove ((obj)->orb->current_invocations, pobj);                        \
-		   --(((ORBit_POAObject)(obj)->adaptor_obj)->use_cnt);                                    \
-		   if (((ORBit_POAObject)(obj)->adaptor_obj)->life_flags & ORBit_LifeF_NeedPostInvoke)    \
-		   ORBit_POAObject_post_invoke (((ORBit_POAObject)(obj)->adaptor_obj));           \
+		   (obj)->orb->current_invocations =
+		   g_slist_remove ((obj)->orb->current_invocations, pobj);
+		   POA_LOCK (((ORBit_POAObject)(obj)->adaptor_obj)->poa);
+		   --(((ORBit_POAObject)(obj)->adaptor_obj)->use_cnt);
+		   if (((ORBit_POAObject)(obj)->adaptor_obj)->life_flags & ORBit_LifeF_NeedPostInvoke)
+			ORBit_POAObject_post_invoke (((ORBit_POAObject)(obj)->adaptor_obj));
+		   POA_UNLOCK (((ORBit_POAObject)(obj)->adaptor_obj)->poa);
 		*/
+
 	} else
 		ORBit_small_invoke_stub_n
 			(obj, methods, method_index,
